#include "cmaudio.h"

// Native audio bus master control registers (offsets)
#define INTERRUPT	0x00 // Interrupt for AC97 controller
#define INTERRUPT_EN	0x04 // Interrupt Enable
#define COMMAND		0x08 // AC'97 Command
#define	PHY_OUT_STATUS	0x0C // Status and address for Phy registers (slot 1) & data
#define	PHY_IN_STATUS	0x10 // Input Phy address from Codec and data
#define	SLOT_REQ	0x14 // SLOTREQ flag bit for next output slot 3-12
#define	SLOT_COUNT	0x18 // Slot counter (0-12) and Bit_Clock counter
#define	BDBAR_PCM_IN	0x1C // Input DMA channel
#define	BDBAR_PCM_OUT	0x34 // Output DMA channel
#define	SPDIF_COMMAND	0x4C // SPDIF Command
#define	MODEM_MIRROR	0x7C // data written by modem
#define	AUDIO_MIRROR	0x80 // data written by audio
#define	REORDER_6CH	0x84 // enable reording of audio output channels
#define	FIFO_FLUSH	0x88 // flush audio FIFO

#define	OFF_THRESHOLD	0x00 // offset to threshold of FIFO and slot enable
#define	OFF_LINK_PTR	0x04 // offset to pointer to link list
#define	OFF_DT_START	0x08 // offset to start of data of current DT
#define	OFF_DT_NEXT	0x0C // pointer to next DT
#define	OFF_DT_CURR	0x10 // pointer to currently accessed memory address
#define	OFF_DT_FIFO	0x14 // size of data block for the current DT, number of FIFO used and free

// Interrupt
typedef struct tagATI_INTERRUPT
{
    u32 DmaInOverFlow     :  1; // 00 (RW) PCM Input Channel DMA Overflow Interrupt
    u32 DmaInStatus       :  1; // 01 (RW) PCM Input Channel DMA Status Interrupt
    u32 DmaOutUnderFlow   :  1; // 02 (RW) PCM Output Channel DMA Underflow Interrupt
    u32 DmaOutStatus      :  1; // 03 (RW) PCM Output Channel DMA Status Interrupt
    u32 DmaSpdifUnderFlow :  1; // 04 (RW) S/PDIF DMA Underflow Interrupt
    u32 DmaSpdifStatus    :  1; // 05 (RW) S/PDIF DMA Status Interrupt
    u32 Reserved_06       :  2; // 06-07
    u32 PhyIn             :  1; // 08 (RW) Got Physical Register Data from Codec Interrupt
    u32 PhyAddrMismatch   :  1; // 09 (RW) Physical Address in/out Mismatch
    u32 NotReadyCodec0    :  1; // 10 (RW) Codec 0 not ready
    u32 NotReadyCodec1    :  1; // 11 (RW) Codec 1 not ready
    u32 NotReadyCodec2    :  1; // 12 (RW) Codec 2 not ready
    u32 NewFrameStart     :  1; // 13 (RO) New Frame Start
    u32 BusBusyAudio      :  1; // 14 (WO) Set Bus Busy Audio
    u32 AudioGpio         :  1; // 15 (RW) Audio Gpio Interrupt (SB300 only)
    u32 Reserved_16       : 16; // 16-31
} ATI_INTERRUPT, *PATI_INTERRUPT;

typedef struct tagATI_COMMAND
{
    u32 PowerDown         :  1; // 00 (RW) Power down
    u32 EnablePcmIn       :  1; // 01 (RW) Enable PCM Input Channel
    u32 EnablePcmOut      :  1; // 02 (RW) Enable PCM Output Channel
    u32 StatusMem         :  1; // 03 (RW) PCM Sending Status
    u32 EnableSpdif       :  1; // 04 (RW) Enable S/PDIF Channel
    u32 StatusSpdif       :  1; // 05 (RW) S/PDIF DMA Sending Status
    u32 ThresholdSpdif    :  2; // 06-07 (RW) Threshold Value for the S/PDIF DMA FIFOs
    u32 EnableDmaPcmIn    :  1; // 08 (RW) Enable PCM Input Channel DMA
    u32 EnableDmaPcmOut   :  1; // 09 (RW) Enable PCM Output Channel DMA
    u32 EnableDmaSpdif    :  1; // 10 (RW) Enable S/PDIF DMA
    u32 StopSpdif         :  1; // 11 (RO) Stop S/PDIF DMA
    u32 SpdifConcurrency  :  3; // 12-14 (RW) S/PDIF Concurrency, S/PDIF Data go to S/PDIF bus
    u32 Reserved_15       :  1; // 15
    u32 InterleaveSpdif   :  1; // 16 (RW) Interleave Format for the S/PDIF Slot
    u32 Reserved_17       :  3; // 17-19
    u32 PresentAudio      :  1; // 20 (RO) Is Audio Present
    u32 InterleaveInput   :  1; // 21 (RW) Interleave Format for the Input Slot
    u32 InterleaveOutput  :  1; // 22 (RW) Interleave Format for the Output Slot
    u32 EnableLoopBack    :  1; // 23 (RW) Enable Loop-back Mode
    u32 DisablePackedFmt  :  1; // 24 (RW) Disable Packed Format
    u32 EnableBurst       :  1; // 25 (RW) Enable Burst Mode
    u32 EnablePanic       :  1; // 26 (RW) Enable Panic Signal
    u32 PresentModem      :  1; // 27 (RO) Is Modem Present
    u32 ACLinkActive      :  1; // 28 (RO) Is AC-Link Active State
    u32 AC97SoftwareReset :  1; // 29 (RO) Assert AC'97 Software Reset
    u32 AC97Sync          :  1; // 30 (RO) Assert AC'97 Sync.
    u32 AC97Reset         :  1; // 31 (RO) De-assert AC'97 Reset
} ATI_COMMAND, *PATI_COMMAND;


typedef struct tagATI_CODEC
{
    u32 CodecID           :  2; // 00-01 (RW) Codec ID
    u32 IsRead            :  1; // 02 (RW) Is Read (1:read,   0:write)
    u32 Reserved_03       :  5; // 03-07
    u32 PhyAddrEnable     :  1; // 08 (RW) Enable Send-out/Receive-from of Physical Address and Data with Next Frame
    u32 PhyAddr           :  7; // 09-15 (RW) Physical in/out Address
    u32 PhyData           : 16; // 16-31 (RW) Physical in/out Data
} ATI_CODEC, *PATI_CODEC;

typedef struct tagATI_SLOT
{
    u32 Slot03            :  1; // 00 (RO) Slot Req Flag bit for Next Output Slot 03
    u32 Slot04            :  1; // 01 (RO) Slot Req Flag bit for Next Output Slot 04
    u32 Slot05            :  1; // 02 (RO) Slot Req Flag bit for Next Output Slot 05
    u32 Slot06            :  1; // 03 (RO) Slot Req Flag bit for Next Output Slot 06
    u32 Slot07            :  1; // 04 (RO) Slot Req Flag bit for Next Output Slot 07
    u32 Slot08            :  1; // 05 (RO) Slot Req Flag bit for Next Output Slot 08
    u32 Slot09            :  1; // 06 (RO) Slot Req Flag bit for Next Output Slot 09
    u32 Slot10            :  1; // 07 (RO) Slot Req Flag bit for Next Output Slot 10
    u32 Slot11            :  1; // 08 (RO) Slot Req Flag bit for Next Output Slot 11
    u32 Slot12            :  1; // 09 (RO) Slot Req Flag bit for Next Output Slot 12
    u32 Reserved_10       :  1; // 10
    u32 Threshold         :  5; // 11-15 (RW) Threshold
    u32 Reserved_16       : 16; // 16-31
} ATI_SLOT, *PATI_SLOT;

typedef struct tagATI_THRESHOLD_IN
{
    u32 Threshold         :  5; // 00-04 (RW) Threshold
    u32 Reserved_05       : 27; // 05-31
} ATI_THRESHOLD_IN, *PATI_THRESHOLD_IN;

typedef struct tagATI_DT_FIFO
{
    u32 DTSize            : 16; // 00-15 Data Size of DT for DMA
    u32 FifoUsed          :  5; // 16-20 Number of Filled FIFO Entries of DMA
    u32 FifoFree          :  5; // 21-25 Number of Free FIFO Entries of DMA
    u32 DmaState          :  3; // 26-28 Current State of DMA
    u32 Reserved_29       :  3; // 29-31
} ATI_DT_FIFO, *PATI_DT_FIFO;

typedef struct tagATI_FIFO_FLUSH
{
    u32 FlushOut          :  1; // 00 (RW) Flush Output DMA FIFO
    u32 FlushIn           :  1; // 01 (RW) Flush Intput DMA FIFO
    u32 Reserved_02       : 30; // 02-31
} ATI_FIFO_FLUSH, *PATI_FIFO_FLUSH;

typedef struct sg_item
{   
    u32       dwPtrToPhyAddress;
    u16       wStatus;
    u16       wDTSize;
    u32       dwPtrNextPhyAddress;
} tBDEntry;

/* an instance of the buffer descriptor */
#define SG_LEN 32

int RequestRegion_ATI_IXP(struct cmedia_card *card)
{
	if (request_mem_region(card->opbase, 256, card_names[card->driver_data]) == NULL)
		goto request_error;
	if ((card->reg_virt = ioremap(card->opbase, 256)) == NULL)
		release_mem_region(card->opbase, 256);
	else
		return 1;
request_error:
	return 0;
}

static void ReleaseRegion_ATI_IXP(struct cmedia_card *card)
{
	if (card->pcmout_buffer_descriptor) {
		pci_free_consistent(card->pci_dev,
			SG_LEN * sizeof(struct sg_item),
			card->pcmout_buffer_descriptor,
	       		card->pcmout_bd_handle);
		card->pcmout_buffer_descriptor = NULL;
	}
	if (card->pcmin_buffer_descriptor) {
		pci_free_consistent(card->pci_dev,
			SG_LEN * sizeof(struct sg_item),
			card->pcmin_buffer_descriptor,
	       		card->pcmin_bd_handle);
		card->pcmin_buffer_descriptor = NULL;
	}
	iounmap(card->reg_virt);
	release_mem_region(card->opbase, 256);
}

static	u32 ReadRegisterDword(struct cmedia_card *card, unsigned offset)
{
	u32 *ptr = (u32 *)(card->reg_virt + offset);
	return *ptr;
}

static	void WriteRegisterDword(struct cmedia_card *card, unsigned offset, u32 data)
{
	u32 *ptr = (u32 *)(card->reg_virt + offset);
	*ptr = data;
}

static u16 ReadRegisterWord_ATI_IXP(struct cmedia_card *card, u8 reg)
{
	u32	dwReg;
	ATI_CODEC *ati_codec = (ATI_CODEC *)&dwReg;

	// write index
	dwReg = ReadRegisterDword(card, PHY_OUT_STATUS);
	udelay(50);
	ati_codec->CodecID = 0;
	ati_codec->IsRead = 1;
	ati_codec->PhyAddrEnable = 1;
	ati_codec->PhyAddr = reg;
	ati_codec->PhyData = 0;
	WriteRegisterDword(card, PHY_OUT_STATUS, dwReg);
	udelay(50);

	// get data
	dwReg = ReadRegisterDword(card, PHY_IN_STATUS);
	return ati_codec->PhyData;
}

static void WriteRegisterWord_ATI_IXP(struct cmedia_card *card, u8 reg, u16 data)
{
	u32	dwReg;
	ATI_CODEC *ati_codec = (ATI_CODEC *)&dwReg;

	// write index
	dwReg = ReadRegisterDword(card, PHY_OUT_STATUS);
	udelay(50);
	ati_codec->CodecID = 0;
	ati_codec->IsRead = 0;
	ati_codec->PhyAddrEnable = 1;
	ati_codec->PhyAddr = reg;
	ati_codec->PhyData = data;
	WriteRegisterDword(card, PHY_OUT_STATUS, dwReg);
	udelay(50);
}

static void acknowledge_interrupt(struct cmedia_card *card)
{
    unsigned long dwReg, dwIntReg;

    return;
    dwIntReg = ReadRegisterDword(card, INTERRUPT_EN);
    
    // disable all interrupts
    dwReg = 0;
    WriteRegisterDword(card, INTERRUPT_EN, dwReg);

    // acknowledge
    dwReg = ReadRegisterDword(card, INTERRUPT);
    if (dwReg)
	    WriteRegisterDword(card, INTERRUPT, dwReg);

    // enable interrupt
    WriteRegisterDword(card, INTERRUPT_EN, dwIntReg);
}

static void check_bus_busy(struct cmedia_card *card)
{
	u32 dwReg;
	ATI_COMMAND *pAtiCmd = (ATI_COMMAND *)&dwReg;
	ATI_INTERRUPT *pAtiInt = (ATI_INTERRUPT *)&dwReg;
	int busy;


	dwReg = ReadRegisterDword(card, COMMAND);
	if (pAtiCmd->EnablePcmIn
	 || pAtiCmd->EnablePcmOut
	 || pAtiCmd->EnableSpdif)
		busy = 1;
	else
		busy = 0;
	dwReg = ReadRegisterDword(card, INTERRUPT_EN);
	pAtiInt->BusBusyAudio = busy;
	WriteRegisterDword(card, INTERRUPT_EN, dwReg);
}

static void StartADC_ATI_IXP(struct cmedia_card *card)
{
	unsigned long flags;
	unsigned long dwReg;
	ATI_COMMAND *command = (ATI_COMMAND *)&dwReg;

	if (!(card->dma_state & ADC_RUNNING)) {
		spin_lock_irqsave(&card->lock, flags);
		card->dma_state |= ADC_RUNNING;
		dwReg = ReadRegisterDword(card, COMMAND);
		command->EnablePcmIn = 1;
		command->EnableDmaPcmIn = 1;
		command->InterleaveInput = 1;
		command->EnableBurst = 1;
		WriteRegisterDword(card, COMMAND, dwReg);
		check_bus_busy(card);
		// link list regster
		WriteRegisterDword(card, BDBAR_PCM_IN + OFF_LINK_PTR,
			virt_to_bus(card->pcmin_buffer_descriptor) | 1);	
		spin_unlock_irqrestore(&card->lock, flags);
	}
}

/* Stop recording (lock held) */
static void StopADC_ATI_IXP(struct cmedia_card *card)
{
	unsigned long dwReg;
	ATI_COMMAND *command = (ATI_COMMAND *)&dwReg;
	ATI_FIFO_FLUSH *flush = (ATI_FIFO_FLUSH *)&dwReg;

	card->dma_state &= ~ADC_RUNNING;
	acknowledge_interrupt(card);
	WriteRegisterDword(card, BDBAR_PCM_IN + OFF_LINK_PTR, 0);
	dwReg = ReadRegisterDword(card, COMMAND);
	command->EnablePcmIn = 0;
	command->EnableDmaPcmIn = 0;
	WriteRegisterDword(card, COMMAND, dwReg);
	check_bus_busy(card);

	dwReg = ReadRegisterDword(card, FIFO_FLUSH);
	flush->FlushIn = 1;
	WriteRegisterDword(card, FIFO_FLUSH, dwReg);
	check_bus_busy(card);
}

static void StartDAC_ATI_IXP(struct cmedia_card *card)
{
	unsigned long flags;
	unsigned long dwReg;
	ATI_COMMAND *command = (ATI_COMMAND *)&dwReg;

	if (!(card->dma_state & DAC_RUNNING)) {
		spin_lock_irqsave(&card->lock, flags);
		card->dma_state |= DAC_RUNNING;
		dwReg = ReadRegisterDword(card, COMMAND);
		command->EnablePcmOut = 1;
		command->EnableDmaPcmOut = 1;
		command->InterleaveOutput = 1;
		command->EnableBurst = 1;
		WriteRegisterDword(card, COMMAND, dwReg);
		check_bus_busy(card);
		// link list regster
		WriteRegisterDword(card, BDBAR_PCM_OUT + OFF_LINK_PTR,
			virt_to_bus(card->pcmout_buffer_descriptor) | 1);
		spin_unlock_irqrestore(&card->lock, flags);
	}
	card->SetMonitorSPDIF(card, 1);
}

/* stop playback (lock held) */
static void __StopDAC_ATI_IXP(struct cmedia_card *card)
{
	unsigned long dwReg;
	ATI_COMMAND *command = (ATI_COMMAND *)&dwReg;
	ATI_FIFO_FLUSH *flush = (ATI_FIFO_FLUSH *)&dwReg;

	card->dma_state &= ~DAC_RUNNING;
 	acknowledge_interrupt(card);
	WriteRegisterDword(card, BDBAR_PCM_OUT + OFF_LINK_PTR, 0);
	dwReg = ReadRegisterDword(card, COMMAND);
	command->EnablePcmOut = 0;
	command->EnableDmaPcmOut = 0;
	WriteRegisterDword(card, COMMAND, dwReg);
	check_bus_busy(card);

	dwReg = ReadRegisterDword(card, FIFO_FLUSH);
	flush->FlushOut = 1;
	WriteRegisterDword(card, FIFO_FLUSH, dwReg);
	check_bus_busy(card);
	card->SetMonitorSPDIF(card, 0);
	
	memset(card->pcmout_rawbuf, 0, PCMOUT_DMABUF_SIZE);
	memset(card->dmaout_rawbuf, 0, PCMOUT_DMABUF_SIZE);
}

/* update buffer manangement pointers */
static void UpdatePtr_ATI_IXP(struct cmedia_card *card)
{
    unsigned long dwReg, buffer_size;

    if (!card->dma_state)
	    return;

    if (card->dma_state & ADC_RUNNING) {
	buffer_size = PCMIN_DMABUF_SIZE;
	dwReg = ReadRegisterDword(card, BDBAR_PCM_IN + OFF_DT_CURR);
	if (dwReg < virt_to_bus(card->pcmin_rawbuf))
		card->pcmin_hwptr = 0;
	else {
		card->pcmin_hwptr = dwReg - virt_to_bus(card->pcmin_rawbuf);
		if (card->pcmin_hwptr >= buffer_size)
			card->pcmin_hwptr = 0;
	}
    }

    if (card->dma_state & DAC_RUNNING) {
	buffer_size = PCMOUT_DMABUF_SIZE * card->output_channels / 6;
	dwReg = ReadRegisterDword(card, BDBAR_PCM_OUT + OFF_DT_CURR);
	if (dwReg < virt_to_bus(card->dmaout_rawbuf))
		card->pcmout_hwptr = 0;
	else {
		card->pcmout_hwptr = dwReg - virt_to_bus(card->dmaout_rawbuf);
		if (card->pcmout_hwptr >= buffer_size)
			card->pcmout_hwptr = 0;
	}
	if (card->output_channels == 2) {
		card->pcmout_hwptr -= card->pcmout_hwptr % 4;
		card->pcmout_hwptr = card->pcmout_hwptr * 3;
	} else if (card->output_channels == 4) {
		card->pcmout_hwptr -= card->pcmout_hwptr % 8;
		card->pcmout_hwptr = card->pcmout_hwptr * 3 / 2;
	} else if (card->output_channels == 6)
		card->pcmout_hwptr -= card->pcmout_hwptr % 12;
    }
}

static int ProgramDMABuffer_Ati_IXP(struct cmedia_card *card)
{
    struct sg_item *sg;
    unsigned long buffer_size, dwReg;
    int i;
    u32 busaddr, slice_size;
    ATI_SLOT *pAtiSlot = (ATI_SLOT *)&dwReg;

    card->pcmout_count = card->pcmin_count = 0;
    card->pcmout_swptr = card->pcmout_hwptr = card->pcmin_swptr = card->pcmin_hwptr = 0;

    memset(card->pcmout_rawbuf, 0, PCMOUT_DMABUF_SIZE);
    memset(card->dmaout_rawbuf, 0, PCMOUT_DMABUF_SIZE);
    memset(card->pcmin_rawbuf, 0, PCMIN_DMABUF_SIZE);

    sg = (struct sg_item *)card->pcmout_buffer_descriptor;
    busaddr = virt_to_bus(card->dmaout_rawbuf);
    buffer_size = PCMOUT_DMABUF_SIZE * card->output_channels / 6;
    slice_size = buffer_size / SG_LEN;
    for (i = 0; i < SG_LEN; i++) {
	sg->dwPtrToPhyAddress = busaddr + slice_size * i;
	sg->wStatus = 0;
	sg->wDTSize = slice_size / 4;
	sg->dwPtrNextPhyAddress = virt_to_bus(sg + 1);
	sg++;
    }
    sg--;
    // last entry
    sg->dwPtrNextPhyAddress = virt_to_bus(card->pcmout_buffer_descriptor);

   card->pcmout_count = 0;

    // -------------------------------------------------------------------------------------------------
    sg = (struct sg_item *)card->pcmin_buffer_descriptor;
    busaddr = virt_to_bus(card->pcmin_rawbuf);
    buffer_size = PCMIN_DMABUF_SIZE;
    slice_size = buffer_size / SG_LEN;
    for (i = 0; i < SG_LEN; i++) {
	sg->dwPtrToPhyAddress = busaddr + slice_size * i;
	sg->wStatus = 0;
	sg->wDTSize = slice_size / 4;
	sg->dwPtrNextPhyAddress = virt_to_bus(sg + 1);
	sg++;
    }
    sg--;
    // last entry
    sg->dwPtrNextPhyAddress = virt_to_bus(card->pcmin_buffer_descriptor);

    card->pcmin_count = 0;

	// Set slot enable
	dwReg = ReadRegisterDword(card, BDBAR_PCM_OUT + OFF_THRESHOLD);
	pAtiSlot->Slot03 =
	pAtiSlot->Slot04 =
	pAtiSlot->Slot06 =
	pAtiSlot->Slot09 =
	pAtiSlot->Slot07 =
	pAtiSlot->Slot08 =
	pAtiSlot->Slot10 =
	pAtiSlot->Slot11 = 0;
	pAtiSlot->Threshold = 4;
	switch (card->output_channels) {
	case 8:
		pAtiSlot->Slot10 =
		pAtiSlot->Slot11 = 1;
	case 6:
		pAtiSlot->Slot07 =
		pAtiSlot->Slot08 = 1;
	case 4:
		pAtiSlot->Slot06 =
		pAtiSlot->Slot09 = 1;
	default:
		pAtiSlot->Slot03 =
		pAtiSlot->Slot04 = 1;
		break;
	}
	WriteRegisterDword(card, BDBAR_PCM_OUT + OFF_THRESHOLD, dwReg);
    return 0;
}

static int ProgramDMABuffer_ATI_IXP(struct cmedia_card *card)
{
    unsigned long flags;
    int ret;
    spin_lock_irqsave(&card->lock, flags);
    ret = ProgramDMABuffer_Ati_IXP(card);
    spin_unlock_irqrestore(&card->lock, flags);
    return ret;
}

static int DrainDAC_ATI_IXP(struct cmedia_card *card)
{
    if (!(card->dma_state & DAC_RUNNING))
	    return 0;

    __StopDAC_ATI_IXP(card);

    return 0;
}

static void ConvertPCMOUT_ATI_IXP(struct cmedia_card *card)
{
    short *ptrSource, *ptrTarget;
    unsigned long i, len1, len2;

    len1 = len2 = 0;

    if (card->pcmout_swptr < card->pcmout_hwptr)
	len1 = ((card->pcmout_hwptr - card->pcmout_swptr) >> 1) / 6;
    else {
	len1 = ((PCMOUT_DMABUF_SIZE - card->pcmout_swptr) >> 1) / 6;
	len2 = ((card->pcmout_hwptr) >> 1) / 6;
    }

    switch (card->output_channels) {
    case 2:
	// Front Left/Right
	ptrSource = (short *)(card->pcmout_rawbuf + card->pcmout_swptr);
	ptrTarget = (short *)(card->dmaout_rawbuf + card->pcmout_swptr / 3);
	for (i = 0; i < len1; i++) {
	    ptrTarget[0] = ptrSource[0];
	    ptrTarget[1] = ptrSource[1];
	    ptrSource += 6;
	    ptrTarget += 2;
	}

	ptrSource = (short *)(card->pcmout_rawbuf);
	ptrTarget = (short *)(card->dmaout_rawbuf);
	for (i = 0; i < len2; i++) {
	    ptrTarget[0] = ptrSource[0];
	    ptrTarget[1] = ptrSource[1];
	    ptrSource += 6;
	    ptrTarget += 2;
	}
	break;

    case 4:
	// Front Left/Right, Rear Left/Right
	ptrSource = (short *)(card->pcmout_rawbuf + card->pcmout_swptr);
	ptrTarget = (short *)(card->dmaout_rawbuf + card->pcmout_swptr * 2 / 3);
	if (card->XearMode) {
	    for (i = 0; i < len1; i++) {
		ptrTarget[2] = ptrSource[0];
		ptrTarget[3] = ptrSource[1];
		ptrTarget[0] = ptrSource[2];
		ptrTarget[1] = ptrSource[3];
	    	ptrSource += 6;
	    	ptrTarget += 4;
	    }
	} else {
	    for (i = 0; i < len1; i++) {
		ptrTarget[0] = ptrSource[0];
		ptrTarget[1] = ptrSource[1];
		ptrTarget[2] = ptrSource[2];
		ptrTarget[3] = ptrSource[3];
	    	ptrSource += 6;
	    	ptrTarget += 4;
	    }
	}

	ptrSource = (short *)(card->pcmout_rawbuf);
	ptrTarget = (short *)(card->dmaout_rawbuf);
	if (card->XearMode) {
	    for (i = 0; i < len2; i++) {
		ptrTarget[2] = ptrSource[0];
		ptrTarget[3] = ptrSource[1];
		ptrTarget[0] = ptrSource[2];
		ptrTarget[1] = ptrSource[3];
	    	ptrSource += 6;
	    	ptrTarget += 4;
	    }
	} else {
	    for (i = 0; i < len2; i++) {
		ptrTarget[0] = ptrSource[0];
		ptrTarget[1] = ptrSource[1];
		ptrTarget[2] = ptrSource[2];
		ptrTarget[3] = ptrSource[3];
	    	ptrSource += 6;
	    	ptrTarget += 4;
	    }
	}
	break;

    case 6:
	// Front Left/Right, Rear Left/Right, Center/Bass
	ptrSource = (short *)(card->pcmout_rawbuf + card->pcmout_swptr);
	ptrTarget = (short *)(card->dmaout_rawbuf + card->pcmout_swptr);
	if (card->XearMode) {
	    for (i = 0; i < len1; i++) {
		ptrTarget[4] = ptrSource[0];
		ptrTarget[5] = ptrSource[1];
		ptrTarget[0] = ptrSource[2];
		ptrTarget[1] = ptrSource[3];
		ptrTarget[2] = ptrSource[4];
		ptrTarget[3] = ptrSource[5];
	    	ptrSource += 6;
	    	ptrTarget += 6;
	    }
	} else {
	    for (i = 0; i < len1; i++) {
		ptrTarget[0] = ptrSource[0];
		ptrTarget[1] = ptrSource[1];
		ptrTarget[4] = ptrSource[2];
		ptrTarget[5] = ptrSource[3];
		ptrTarget[2] = ptrSource[4];
		ptrTarget[3] = ptrSource[5];
	    	ptrSource += 6;
	    	ptrTarget += 6;
	    }
	}

	ptrSource = (short *)(card->pcmout_rawbuf);
	ptrTarget = (short *)(card->dmaout_rawbuf);
	if (card->XearMode) {
	    for (i = 0; i < len2; i++) {
		ptrTarget[4] = ptrSource[0];
		ptrTarget[5] = ptrSource[1];
		ptrTarget[0] = ptrSource[2];
		ptrTarget[1] = ptrSource[3];
		ptrTarget[2] = ptrSource[4];
		ptrTarget[3] = ptrSource[5];
	    	ptrSource += 6;
	    	ptrTarget += 6;
	    }
	} else {
	    for (i = 0; i < len2; i++) {
		ptrTarget[0] = ptrSource[0];
		ptrTarget[1] = ptrSource[1];
		ptrTarget[4] = ptrSource[2];
		ptrTarget[5] = ptrSource[3];
		ptrTarget[2] = ptrSource[4];
		ptrTarget[3] = ptrSource[5];
	    	ptrSource += 6;
	    	ptrTarget += 6;
	    }
	}
	break;
    }
}

int Initial_ATI_IXP(struct cmedia_card *card)
{
    u16 eid;
    int i = 0;
    u32 dwReg;
    ATI_COMMAND *pAtiCommand = (ATI_COMMAND *)&dwReg;

	card->ReleaseRegion = ReleaseRegion_ATI_IXP;
	card->ReadRegisterWord = ReadRegisterWord_ATI_IXP;
	card->WriteRegisterWord = WriteRegisterWord_ATI_IXP;
	card->StartADC = StartADC_ATI_IXP;
	card->StopADC = StopADC_ATI_IXP;
	card->StartDAC = StartDAC_ATI_IXP;
	card->DrainDAC = DrainDAC_ATI_IXP;
	card->UpdatePtr = UpdatePtr_ATI_IXP;
	card->ProgramDMABuffer = ProgramDMABuffer_ATI_IXP;
	card->ConvertPCMOUT = ConvertPCMOUT_ATI_IXP;

	if ((card->pcmout_buffer_descriptor = pci_alloc_consistent(card->pci_dev,
		SG_LEN * sizeof(struct sg_item),
		&card->pcmout_bd_handle)) == NULL) {
		printk(KERN_ERR "CMedia Audio: out of memory\n");
		return -ENOMEM;
	}
	if ((card->pcmin_buffer_descriptor = pci_alloc_consistent(card->pci_dev,
		SG_LEN * sizeof(struct sg_item),
		&card->pcmin_bd_handle)) == NULL) {
		// free pcmout buffer descriptor
		pci_free_consistent(card->pci_dev,
			SG_LEN * sizeof(struct sg_item),
			card->pcmout_buffer_descriptor,
	       		card->pcmout_bd_handle);
		card->pcmout_buffer_descriptor = NULL;
		printk(KERN_ERR "CMedia Audio: out of memory\n");
		return -ENOMEM;
	}

	// reset powerdown
	dwReg = ReadRegisterDword(card, COMMAND);
	pAtiCommand->PowerDown = 0;
	WriteRegisterDword(card, COMMAND, dwReg);
	udelay(10);

	// soft reset
	dwReg = ReadRegisterDword(card, COMMAND);
	pAtiCommand->AC97SoftwareReset = 1;
	WriteRegisterDword(card, COMMAND, dwReg);
	udelay(10);
	dwReg = ReadRegisterDword(card, COMMAND);
	pAtiCommand->AC97SoftwareReset = 0;
	WriteRegisterDword(card, COMMAND, dwReg);

	while ((dwReg = ReadRegisterDword(card, COMMAND))
		&& !pAtiCommand->ACLinkActive) {
		// hard reset
		pAtiCommand->AC97Sync = 1;
		pAtiCommand->AC97Reset = 0;
		WriteRegisterDword(card, COMMAND, dwReg);
		dwReg = ReadRegisterDword(card, COMMAND);
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(1);
		pAtiCommand->AC97Reset = 1;
		WriteRegisterDword(card, COMMAND, dwReg);
	}

    /* power up everything, modify this when implementing power saving */
    card->WriteRegisterWord(card, AC97_POWER_CONTROL,
	card->ReadRegisterWord(card, AC97_POWER_CONTROL) & ~0x7f00);

    /* wait for analog ready */
    for (i = 10; i && ((card->ReadRegisterWord(card, AC97_POWER_CONTROL) & 0xf) != 0xf); i--) {
	current->state = TASK_UNINTERRUPTIBLE;
	schedule_timeout(card->process_time / 20);
    }

    /* Don't attempt to get eid until powerup is complete */
    eid = card->ReadRegisterWord(card, AC97_EXTENDED_ID);

    if (eid == 0xFFFF) {
	printk(KERN_WARNING "CMedia Audio: no codec attached ?\n");
	return 0;
    }

    card->ac97_features = eid;

    /* Now check the codec for useful features to make up for
       the dumbness of the 810 hardware engine */

    if (eid & 0x0001) {
	printk(KERN_WARNING "CMedia Audio: Codec refused to allow VRA, using 48Khz only.\n");
	card->ac97_features &= ~1;
    }

    card->ac97_vendor_id1 = card->ReadRegisterWord(card, AC97_VENDOR_ID1);
    card->ac97_vendor_id2 = card->ReadRegisterWord(card, AC97_VENDOR_ID2);

    if (card->ac97_vendor_id1 != 0x434d) {
	printk(KERN_ERR "CMedia Audio: CMedia codec not found.\n");
	return 0;
    }

    return 1;
}
